/**
 * @typedef {object} Options
 * @property {boolean} ignoreLeadingInlineMemberAccess
 * @property {number} ignoreTopChainWithDepth
 * @property {boolean} includeMemberAccess
 * @property {'always' | 'never' | 'only-single-chain'} leadingNewlineForTopChain
 */

/**
 * @typedef {import('eslint').AST.Token} Token
 * @typedef {import('estree').Node} AllNode
 * @typedef {import('estree').Node['type']} AllNodeType
 * @typedef {import('estree').BaseNodeWithoutComments} Base
 * @typedef {import('estree').Comment} Comment
 * @typedef {import('eslint').Rule.NodeParentExtension} ParentExtension
 * @typedef {import('estree').MemberExpression} MemberExpression
 * @typedef {import('eslint').SourceCode} SourceCode
 */

/**
 * @template N
 * @typedef {N & ParentExtension} NodeWithParent
 */

/**
 * main idea from `eslint/lib/rules/utils/ast-utils`
 */
const astUtils = {
	/**
	 * @param {Token} token
	 */
	isComment: (token) =>
		[ 'Line', 'Block', 'Shebang' ].includes(token.type),
	/**
	 * @param {Token | AllNode | Comment} left
	 * @param {Token | AllNode | Comment} right
	 */
	isOnSameLine: (left, right) =>
		left.loc.end.line === right.loc.start.line,
}
/**
 * @template T
 * @template U
 * @template V
 * @param {(x: T) => U} f
 * @param {(x: U) => V} g
 */
const compose = (g, f) =>
	/**
	 * @param {T} x
	 */
	(x) => g(f(x))
/**
 * @template {AllNodeType} K
 * @template {Extract<AllNode, { type: K }>} N
 * @param {K} type
 * @param {keyof NodeWithParent<N>} propName
 */
const skip = (type, propName, shouldEqual = true) =>
	/**
	 * @param {AllNode} node
	 * @returns {any}
	 */
	(node) =>
		(shouldEqual
			? node?.type === type
			: node?.type !== type
		)
			? node[propName]
			: node
const skipCallExpression = skip('CallExpression', 'callee')
const skipChainExpression = skip('ChainExpression', 'expression')
const skipCallAndChain = compose(skipChainExpression, skipCallExpression)
/**
 * @param {AllNode['type']} type
 */
const skipParent = (type, shouldEqual = true) =>
	skip(type, 'parent', shouldEqual)
const skipParentCallExpression = skipParent('CallExpression')
const skipParentChainExpression = skipParent('ChainExpression')
const skipParentChainAndParentCall = compose(skipParentCallExpression, skipParentChainExpression)

/**
 * @param {NodeWithParent<MemberExpression>} memberNode
 */
const isMethodCall = (memberNode) =>
	skipParentChainExpression(memberNode.parent).type === 'CallExpression'
/**
 * @param {NodeWithParent<MemberExpression>} memberNode
 */
const isFirstTopChain = (memberNode) =>
	memberNode?.object && skipCallAndChain(memberNode.object).type !== 'MemberExpression'

/**
 * @param {SourceCode} sourceCode
 */
 const takePrevTokenOnSourceCode = (sourceCode) =>
	/**
	 * @param {MemberExpression} node
	 */
	(node, includeComments = true) =>
		sourceCode.getTokenBefore(node, { includeComments })
/**
 * @param {SourceCode} sourceCode
 */
const takePropertyAccessorOnSourceCode = (sourceCode) =>
	/**
	 * @param {MemberExpression} node
	 */
	(node, includeComments = true) =>
		takePrevTokenOnSourceCode(sourceCode)(node.property, includeComments)
/**
 * @param {NodeWithParent<AllNode>} node
 * @param {Options} options
 */
const takeNextChain = (node, options) => {
	if (options.includeMemberAccess) return skipParentChainAndParentCall(node)
	
	let temp = node
	
	while (true) {
		temp = skipParentChainExpression(temp.parent)
		const { parent } = temp
		
		if (parent.type === 'CallExpression') return skipParentCallExpression(temp)
		else if (parent.type !== 'MemberExpression') return null
	}
}
/**
 * @param {NodeWithParent<MemberExpression>} firstTopChainNode
 * @param {SourceCode} sourceCode
 * @param {Options} options
 */
const takeTopChains = (firstTopChainNode, sourceCode, options) => {
	const takePropertyAccessor = takePropertyAccessorOnSourceCode(sourceCode)
	const topChains = []
	let child
	let temp
	let foundMethodCall = false
	
	while (true) {
		temp = child ? takeNextChain(child.parent, options) : firstTopChainNode
		
		if (temp?.type !== 'MemberExpression') break
		
		const propertyAccessorToken = takePropertyAccessor(firstTopChainNode, false)
		const tempPropertyAccessorToken = takePropertyAccessor(temp, false)
		
		if (!astUtils.isOnSameLine(propertyAccessorToken, tempPropertyAccessorToken)) break
		
		child = temp
		
		if (options.ignoreLeadingInlineMemberAccess) {
			foundMethodCall ||= isMethodCall(child)
			if (!foundMethodCall) continue
		}
		topChains.push(child)
	}
	
	return topChains
}

/**
 * @type {import('eslint').Rule.RuleModule}
 */
module.exports = {
	meta: {
		type: 'layout',
		fixable: 'whitespace',
		schema: [
			{
				type: 'object',
				properties: {
					ignoreLeadingInlineMemberAccess: {
						type: 'boolean',
						default: true,
					},
					ignoreTopChainWithDepth: {
						type: 'integer',
						minimum: 1,
						default: 1,
					},
					includeMemberAccess: {
						type: 'boolean',
						default: true,
					},
					leadingNewlineForTopChain: {
						enum: [ 'always', 'never', 'only-single-chain' ],
						default: 'only-single-chain',
					},
				},
			}
		],
		messages: {
			unexpectedLineBreak: 'There should be no line break here.',
			missingLineBreak: 'There should be a line break here.',
		},
	},
	create(context) {
		/**
		 * @type {Options}
		 */
		const options = context.options[0] ?? {}
		options.ignoreLeadingInlineMemberAccess ??= true
		options.ignoreTopChainWithDepth ??= 1
		options.includeMemberAccess ??= true
		options.leadingNewlineForTopChain ??= 'only-single-chain'
		
		// ignoreLeadingInlineMemberAccess requires includeMemberAccess to be enabled
		options.ignoreLeadingInlineMemberAccess &&= options.includeMemberAccess
		
		const sourceCode = context.getSourceCode()
		
		const takePrevToken = takePrevTokenOnSourceCode(sourceCode)
		const takePropertyAccessor = takePropertyAccessorOnSourceCode(sourceCode)
		
		/**
		 * @param {string} msgId
		 * @param {(fixer: import('eslint').Rule.RuleFixer, prevToken: Token, token: Token) => import('eslint').Rule.Fix | null} fix
		 */
		const createReporter = (msgId, fix) =>
			/**
			 * @param {Token | AllNode | import('estree').Comment} token
			 */
			(token) => {
				const prevToken = takePrevToken(token)
				
				context.report({
					loc: {
						start: prevToken.loc.end,
						end: token.loc.start,
					},
					messageId: msgId,
					fix: (fixer) => fix(fixer, prevToken, token),
				})
			}
		const reportMissingLineBreak = createReporter(
			'missingLineBreak',
			(fixer, prevToken, token) =>
				fixer.replaceTextRange([ prevToken.range[1], token.range[0] ], '\n')
		)
		const reportUnexpectedLineBreak = createReporter(
			'unexpectedLineBreak',
			(fixer, prevToken, token) => {
				if (astUtils.isComment(prevToken)) return null
				if (astUtils.isOnSameLine(prevToken, token)) return null
				
				return fixer.replaceTextRange([ prevToken.range[1], token.range[0] ], '')
			}
		)
		
		return {
			/**
			 * @param {NodeWithParent<MemberExpression>} node
			 */
			'MemberExpression:exit'(node) {
				const propertyAccessorToken = takePropertyAccessor(node)
				
				const fixChainsOnFirstTopChain = (topChains) => {
					const targetNode = skipChainExpression(node.object)
					const isTargetAndPropertyAccessorOnSameLine = astUtils.isOnSameLine(targetNode, propertyAccessorToken)
					
					if (options.leadingNewlineForTopChain === 'always' && isTargetAndPropertyAccessorOnSameLine) {
						/**
						 * a.b()
						 * .c()
						 * 
						 * should be fixed to
						 * 
						 * a
						 * .b()
						 * .c()
						 */
						reportMissingLineBreak(propertyAccessorToken)
					} else if (options.leadingNewlineForTopChain === 'never' && !isTargetAndPropertyAccessorOnSameLine) {
						/**
						 * a
						 * .b()
						 * .c()
						 * 
						 * should be fixed to
						 * 
						 * a.b()
						 * .c()
						 */
						reportUnexpectedLineBreak(propertyAccessorToken)
					} else if (options.leadingNewlineForTopChain === 'only-single-chain') {
						const isTopChainMultiple = topChains.length > 1
						
						if (isTopChainMultiple) {
							if (!astUtils.isOnSameLine(targetNode, propertyAccessorToken)) {
								/**
								 * a
								 * .b().c()
								 * .d()
								 * 
								 * should be fixed to
								 * 
								 * a.b().c()
								 * .d()
								 */
								reportUnexpectedLineBreak(propertyAccessorToken)
							}
						} else if (topChains.length !== 0) {
							const firstTopChainPropertyAccessor = takePropertyAccessor(topChains[0])
							
							if (astUtils.isOnSameLine(targetNode, firstTopChainPropertyAccessor)) {
								/**
								 * a.b()
								 * .c()
								 * 
								 * should be fixed to
								 * 
								 * a
								 * .b()
								 * .c()
								 */
								reportMissingLineBreak(firstTopChainPropertyAccessor)
							}
						}
					}
					if (options.ignoreTopChainWithDepth < topChains.length) {
						/**
						 * a.b().c()
						 * 
						 * should be fixed to
						 * 
						 * a.b()
						 * .c()
						 * 
						 * when ignoreTopChainWithDepth is set to 1
						 */
						topChains.slice(options.ignoreTopChainWithDepth).forEach(chain => {
							const chainPropertyAccessorToken = takePropertyAccessor(chain)
							reportMissingLineBreak(chainPropertyAccessorToken)
						})
					}
				}
				if (!options.includeMemberAccess && !isMethodCall(node)) {
					if (isFirstTopChain(node)) {
						/**
						 * follows leadingNewlineForTopChain behavior
						 */
						const [ _sameNode, ...topChains ] = takeTopChains(node, sourceCode, options)
						fixChainsOnFirstTopChain(topChains)
					}
					return
				}
				
				if (isFirstTopChain(node)) {
					const topChains = takeTopChains(node, sourceCode, options)
					fixChainsOnFirstTopChain(topChains)
				} else {
					const firstTopChain = (() => {
						let target = node
						
						while (true) {
							target = skipCallAndChain(target.object)
							if (isFirstTopChain(target)) return target
						}
					})()
					
					if (astUtils.isOnSameLine(firstTopChain, propertyAccessorToken)) {
						/**
						 * a.b().c()
						 * 
						 * will be fixed when fixing first top chain,
						 * so there is nothing to do here
						 */
						return
					}
					
					/**
					 * @type {NodeWithParent<MemberExpression> | null}
					 */
					const prevChain = options.includeMemberAccess
						? skipCallAndChain(node.object)
						: (() => {
								let temp = node
								
								while (true) {
									temp = skipCallAndChain(temp.object)
									if (isMethodCall(temp) || temp.type !== 'MemberExpression') return temp
								}
							})()
					
					if (prevChain === null) return
					
					if (astUtils.isOnSameLine(prevChain, propertyAccessorToken)) {
						/**
						 * a
						 * .b()
						 * .c().d()
						 * 
						 * should be fixed to
						 * 
						 * a
						 * .b()
						 * .c()
						 * .d()
						 */ 
						reportMissingLineBreak(propertyAccessorToken)
					}
				}
			},
		}
	}
}
